# -*- coding: UTF-8 -*-
import json
from pandas import DataFrame
from .__sqltool__.action import Action


def _make_func(data):
    if type(data) == str:
        data = data.replace("'", "\\'").strip()
    if type(data) in (tuple, list, dict):
        data = json.dumps(data, ensure_ascii=False)
    return data


# 用于操作的集成数据库类型
class SqlData:
    def __init__(self, data: list[dict] or DataFrame):
        df = data if type(data) == DataFrame else DataFrame(data=data)
        df.rename(columns={lie: lie.strip()
                  for lie in df.columns}, inplace=True)
        self._data = df.where(df.notnull(), None).applymap(_make_func)

    def __bool__(self):
        return len(self._data) > 0

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data

    # 获取
    def getLies(self) -> list:
        return list(self._data.columns)

    def getDf(self) -> DataFrame:
        return self._data

    def getDts(self) -> list[dict]:
        return self._data.to_dict(orient='records')

    def getValues(self, lies: list = None) -> list[tuple]:
        if lies:
            return list(zip(*[self._data[lie].tolist() for lie in lies]))
        else:
            return self._data.to_records(index=False).tolist()

    def getAutoColMap(self, class_map: dict, **colmap) -> dict:
        dt = {}
        # 以首行数据为基准进行自动类型分配
        for lie, value in self._data.iloc[0].to_dict().items():
            if colmap.get(lie):
                dt[lie] = colmap[lie]
                continue
            if type(value) == str:
                if len(lie) < 255:
                    dt[lie] = class_map['varchar']
                else:
                    dt[lie] = class_map['text']
            elif type(value) == bool:
                dt[lie] = class_map['bool']
            elif type(value) == int:
                dt[lie] = class_map['int']
            elif type(value) == float:
                dt[lie] = class_map['float']
            else:
                dt[lie] = class_map['varchar']
        return dt

# 表格去重


def table_drop_duplicates(dts: list, *lies) -> DataFrame:
    df = DataFrame(data=dts)
    df.drop_duplicates(subset=lies, keep='first', inplace=True)
    return df


# 数据保存工具
class DataUpdateSave:
    def __init__(self, sqlt: Action, main_keys: list, iftz=True):
        self.sqlt = sqlt
        self.main_keys = main_keys
        self.iftz = iftz

    def getWhere(self, datadts):
        assert datadts
        wheres = []
        for dt in datadts:
            txt = "','".join([str(dt[k]) for k in self.main_keys])
            wheres.append(f"('{txt}')")
        # 删除主表主键数据
        where = f"({','.join(self.sqlt.getLiestrs(self.main_keys))}) in ({','.join(wheres)})"
        return where

    def getChildDts(self, datadts, chlie, if_pop=True) -> list:
        all_chls = []
        for dt in datadts:
            if if_pop:
                chls = dt.pop(chlie, [])
            else:
                chls = dt.get(chlie, [])
            # 自动转换
            if type(chls) == str:
                chls = json.loads(chls)
            for key in self.main_keys:
                for chdt in chls:
                    chdt[key] = dt[key]
            all_chls.extend(chls)
        return all_chls

    def updateMain(self, datadts: list, table, json_lies: list = (), colmap: dict = {}):
        if not datadts:
            if self.iftz:
                print('没有主数据')
            return False
        where = self.getWhere(datadts)
        if self.sqlt.ifExist(table):
            if self.iftz:
                print('删除主表主键')
            self.sqlt.delete(table, where)
        else:
            if self.iftz:
                print('暂无主表:', table)
        for lie in json_lies:
            colmap[lie] = 'text'
            for dt in datadts:
                dt[lie] = json.dumps(dt.get(lie, ''), ensure_ascii=False)
        # 插入数据
        sqldata = SqlData(datadts)
        self.sqlt.createTable(table, sqldata.getLies(),
                              colmap=colmap, key=self.main_keys)
        self.sqlt.insert(table, sqldata)
        if self.iftz:
            print(table, '主表数据量:', len(sqldata))

    def updateChild(self, datadts, child_table, chlie, if_pop=True, colmap=None):
        chdts = self.getChildDts(datadts, chlie, if_pop=if_pop)
        if chdts:
            where = self.getWhere(datadts)
            if self.sqlt.ifExist(child_table):
                if self.iftz:
                    print('删除子表主键')
                self.sqlt.delete(child_table, where)
            else:
                if self.iftz:
                    print('暂无子表:', child_table)
            # 插入数据
            sqldata = SqlData(datadts)
            self.sqlt.createTable(
                child_table, sqldata.getLies(), colmap=colmap)
            self.sqlt.insert(child_table, *chdts)
            if self.iftz:
                print(child_table, '子表数据量:', len(sqldata))
        else:
            if self.iftz:
                print('没有子表数据!')

    def commit(self):
        self.sqlt.commit()
        if self.iftz:
            print('提交成功!')



# from pymssql import connect as sst_connect
from .__sqltool__.action import Action
from .__sqltool__.body import encrypt, decrypt
from .__sqltool__.alter import Alter
import pyodbc

class Sqlserver(Action, Alter):
    def __init__(self, dbname, user, pwd,
                 host='127.0.0.1', port: int = 1433,
                 charset='UTF8',  # cp936
                 ifencryp=True, custom_db_func = None):
        
        def db_func():
            db = pyodbc.connect(
                f'DRIVER={{ODBC Driver 17 for SQL Server}};SERVER={host};DATABASE={dbname};UID={user};PWD={decrypt(user, pwd) if ifencryp else pwd}')  # DRIVER={SQL Server};
            # db = sst_connect(host=self.host, port=self.port,
            #                      user=self.user, password=self.passwd,
            #                      database=self.dbname, charset=self.charset)
            return db
        
        class_map = {'varchar': 'nvarchar(255)',
                     'text': 'ntext',
                     'bool': 'bit',
                     'int': 'int',
                     'float': 'float'
                     }
        super().__init__(custom_db_func if custom_db_func else db_func, placeholder='?', class_map=class_map)

    def getTablestr(self, table):
        if '.' in table:
            t1, t2 = table.split('.')
            return f'[{t1}].[{t2}]'
        else:
            return f'[{table}]'

    def getLiestr(self, lie):
        return f"[{lie.lower()}]"

    def deleteTable(self, table):
        # DROP TABLE table_name
        sql = '''
            If Exists (select * from sysobjects where id = object_id('{table}') and OBJECTPROPERTY(id, 'IsUserTable') = 1)
                DROP TABLE {table} 
        '''.format(table=self.getTablestr(table))
        return self.run(sql)

    # 判断表是否存在
    def ifExist(self, table):
        return Action.ifExist(self, table, def_schema='dbo')


from .body import Body
import traceback


class Action(Body):
    def createTable(self, table, lies: list,
                    colmap: dict = None,
                    key: list or str = None,  # type: ignore
                    suffix: str = None):
        colmap = colmap if colmap else {}
        assert lies, "数量有误！"
        liestrs = [self.getLiestr(lie) for lie in lies]
        if key:
            if type(key) == str:
                key = [key]
            key = f" ,PRIMARY KEY({','.join(self.getLiestrs(key))})"
        else:
            key = ''
        # 默认类型为短字符串类型
        sql = '''
            Create Table If Not Exists {table}
            ({lies} {key}) {suffix}
        '''.format(table=self.getTablestr(table),
                   lies=",\n".join(f'{liestr} {colmap.get(lie, self.class_map["varchar"])}'
                                   for liestr, lie in zip(liestrs, lies)),
                   key=key,
                   suffix=suffix if suffix else '')
        return self.run(sql)

    # 删除表
    def deleteTable(self, table):
        # DROP TABLE table_name
        sql = f'DROP TABLE IF EXISTS {self.getTablestr(table)}'
        return self.run(sql)

    def _auto_add_lie(self, lies, table, colclass):
        have_lies = self.select('*', table, where='1=0', ifgetlies=True)[0]
        not_liest = set(lies) - set(have_lies)
        for lie in not_liest:
            self.run(
                f'ALTER TABLE {self.getTablestr(table)} ADD {self.getLiestr(lie)} {colclass};')
            print(f'{table} 新增字段 {lie}')

    # 插入
    def insert(self, table: str, lies, datas: list,
               in_hz='',
               auto_lie_class=None):
        assert datas, "插入数据不能为空！"
        # 插入语句
        sql = '''INSERT INTO {table}({liestr})
                VALUES ({cstr}) {in_hz}
        '''.format(table=self.getTablestr(table), liestr=','.join(self.getLiestrs(lies)),
                   cstr=','.join([self.placeholder] * len(lies)), in_hz=in_hz)
        if auto_lie_class:
            try:
                return self.in_run(sql, *datas)
            except:
                self._auto_add_lie(lies, table, auto_lie_class)
                return self.in_run(sql, lies, *datas)
        else:
            return self.in_run(sql, *datas)

    # 插入
    def insert_dts(self, table: str, dts: list[dict],
                   in_hz='',
                   auto_lie_class=None):
        lies = dts[0].keys()
        return self.insert(table, lies, [list(dt.values()) for dt in dts],
                           in_hz=in_hz, auto_lie_class=auto_lie_class)

    def create_insert(self, table, dts: list[dict],
                      colmap: dict = None,
                      key: list or str = None,  # type: ignore
                      hz: str = None,
                      in_hz='',
                      auto_lie_class=None):
        lies = dts[0].keys()
        self.createTable(table, lies, colmap, key=key, hz=hz)
        self.insert(table, lies, [list(dt.values()) for dt in dts],
                    in_hz=in_hz, auto_lie_class=auto_lie_class)

    # 修改
    def update(self, table, dt: dict, where: str):
        assert dt, "数据为空！"
        lies = dt.keys()
        setv = ','.join(
            f"{lie}={self.placeholder}" for lie in self.getLiestrs(lies))
        sql = '''UPDATE {table}
                SET {setv}
                WHERE {where}
        '''.format(table=self.getTablestr(table), setv=setv, where=where)
        return self.in_run(sql, list(dt.values()))

    # 删除
    def delete(self, table, where: str):
        # 表名
        sql = '''DELETE FROM {table}
            WHERE {where}
        '''.format(table=self.getTablestr(table), where=where)
        return self.run(sql)

    def select(self, lies, table, where=None, other='',
               data_class='dts',
               ifget_one_lie=False,
               if_original_table=True,
               if_original_lies=True) -> list:
        if type(lies) == str: lies = [lies]
        sql = '''select {lies}
                   from {table}
                   where {where}
                   {other}
           '''.format(lies=','.join(lies if if_original_lies else self.getLiestrs(lies)),
                      table=table if if_original_table else self.getTablestr(table),
                      where=where if where else '1=1',
                      other=other)
        self.run(sql)
        # 获取数据
        if ifget_one_lie:
            return [d[0] for d in self.getResult(data_class='ls')]
        else:
            return self.getResult(data_class=data_class)

    # 判断是否可以查询到
    def ifGet(self, table, where=None, if_error=True):
        try:
            if len(self.select('1', table, where=where, other='limit 1')) > 0:
                return True
            else:
                return False
        except:
            if if_error:
                traceback.print_exc()
            return False

    # 获取数量
    def getNum(self, table, where=None):
        num = self.select('count(1)', table, where=where, data_class='ls')[0][0]
        return num



from pymysql import connect as myt_connect
from .__sqltool__.body import encrypt, decrypt
from .__sqltool__.action import Action
from .__sqltool__.alter import Alter
import importlib


class Mysql(Action, Alter):
    def __init__(self, dbname, user, pwd,
                 host='127.0.0.1', port: int = 3306,
                 charset='utf8mb4',
                 ifencryp=True, custom_db_func=None):
        def db_func():
            return myt_connect(host=host, port=port,
                               user=user, passwd=decrypt(user, pwd) if ifencryp else pwd,
                               database=dbname, charset=charset)

        class_map = {'varchar': 'varchar(255)',
                     'text': 'text',
                     'bool': 'blob',
                     'int': 'int',
                     'float': 'float'
                     }
        super().__init__(custom_db_func if custom_db_func else db_func,
                         placeholder='%s', class_map=class_map)

    def run(self, sql):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.run(self, sql)

    def in_run(self, sql, *datas, if_error_one=True):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.in_run(self, sql, *datas, if_error_one=if_error_one)

    def commit(self):
        # 设置重连
        self._db.ping(reconnect=True)
        return Action.commit(self)

    def getTablestr(self, table):
        return '`%s`' % table

    def getLiestr(self, lie):
        return f"`{lie.lower()}`"

    def ifExist(self, table):
        b = self.run(f'show tables like "{table}"')
        return b and len(self._cursor.fetchall()) > 0


class MysqlPool:
    def __init__(self, dbname, user, pwd, pool_size=10,
                 host='127.0.0.1', port: int = 3306,
                 charset='utf8mb4',
                 ifencryp=True, **kwargs):
        # 动态加载模块
        # pip install mysql-connector-python
        try:
            module = importlib.import_module('mysql.connector')
        except:
            raise ImportError('需要安装: pip install mysql-connector-python')
        # 使用动态加载的模块
        self.pool = module.pooling.MySQLConnectionPool(pool_name="mysqlpool", pool_size=pool_size,
                                                       host=host, port=port,
                                                       user=user, password=decrypt(user, pwd) if ifencryp else pwd,
                                                       database=dbname, charset=charset, **kwargs)

    def get(self) -> Mysql:
        return Mysql(None, None, None, custom_db_func=self.pool.get_connection)

    def close(self):
        try:
            self.pool.close()
        except:
            pass
